grammar Wip;

@header {
package v7.name.panitz.wip;
import static v7.name.panitz.wip.Tree.*;

import java.util.List;
import java.util.ArrayList;

}

startSymbol returns [Tree result]
    :
        {
            var funs = new ArrayList<FunDef>();
            var cons = new ArrayList<ConDef>();
        }
   	 (con=conDef() { cons.add($con.result);} 
 	 |fun=funDef() { funs.add($fun.result);})*
        {$result = new Program(cons,funs);}
    ;


funDef returns [FunDef result]
    :
        {
            var args = new ArrayList<Param >();   
        }
        
        'fun' n=Ident '('
        ( arg=Ident ':' typ=Ident { args.add(new Param($arg.text,$typ.text));}
	        (',' arg=Ident ':' typ= Ident { args.add(new Param($arg.text,$typ.text));}
            )*
        )? ')' ':' typ=Ident
        '=' body=expr
        {
            $result = new FunDef($n.text,args,$body.result,$typ.text);
        }
    ;


conDef returns [ConDef result]:
        {
            var args = new ArrayList<Tree.Param >();
        }
        'con' n=Ident '(' ( an=Ident  ':' at = Ident {args.add(new Param($an.text,$at.text));}
                          (',' an=Ident ':' at=Ident  {args.add(new Param($an.text,$at.text));} )* )? ')'
        {$result = new ConDef($n.text,args);}
    ;


expr returns [Tree result]:
        {
            var cases = new ArrayList<Tree.Case >();
        }
     'case' e=expr '{' (p=pattern {cases.add($p.result);})+ '}' {$result=new CaseExpr($e.result,cases);}
   | 'if' c=expr 'then' a1=expr 'else' a2=expr {$result=new IfExpr($c.result,$a1.result,$a2.result);}
   | ae=addExpr {$result=$ae.result;} 
;

pattern returns [Case result] :
        'of' c=Ident v=Ident '->'  b=expr
        {
            $result = new Case($c.text,$v.text,$b.result);
        }
    ;

addExpr returns [Tree result]:
{
  boolean isSub= false;
}

	e1=multExpr {$result=$e1.result;}
     (('+'| '-'{isSub=true;})
      e2=multExpr {$result=new OpExpr($e1.result,isSub?Operator.sub:Operator.add,$e1.result);} )*
	
;

multExpr returns [Tree result]:
	e1=atom {$result=$e1.result;} ('*' e2=atom {new OpExpr($e1.result,Tree.Operator.mult,$e2.result);})*
    ;


atom returns [Tree result]:
   il=IntegerLiteral {$result=new IntLiteral(Integer.parseInt($il.text));} 
  | '(' e=expr {$result = $e.result;}')'
  | v=Ident {$result = new Variable($v.text);}
        (
            {
                var params = new ArrayList<Tree >();
            }
            '(' (	p=expr {params.add($p.result);}
                    (',' p=expr {params.add($p.result);})*
                )?
            {$result = new FunCall($v.text,params);}
            ')'
        )?
        ('.' field= Ident  {$result=new FieldSelect($result,$field.text);})*
    ;

Ident 
    :   Nondigit
        (   Nondigit
        |   Digit
        )*
    ;

IntegerLiteral
    :    ('0'..'9')+ ('.' ('0'..'9')+)?
    ;

WS  
    :   (' ' | '\t' | '\r'| '\n') -> channel(HIDDEN)
    ;


fragment
Nondigit
    :   [a-zA-Z_]
    ;

fragment
Digit
    :   [0-9]
    ;
