package v8.name.panitz.wip;

import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

import v8.name.panitz.wip.Tree.Case;
import v8.name.panitz.wip.Tree.ConDef;
import v8.name.panitz.wip.Tree.FieldSelect;
import v8.name.panitz.wip.Tree.FunCall;
import v8.name.panitz.wip.Tree.FunDef;
import v8.name.panitz.wip.Tree.IfExpr;
import v8.name.panitz.wip.Tree.IntLiteral;
import v8.name.panitz.wip.Tree.OpExpr;
import v8.name.panitz.wip.Tree.Param;
import v8.name.panitz.wip.Tree.Program;
import v8.name.panitz.wip.Tree.SwitchExpr;
import v8.name.panitz.wip.Tree.Variable;

public class GenCode implements Visitor<Void> {
	Writer w;
	Map<String, Integer> vars;

	int label = 0;

	String nextLabel() {
		return ".L" + (label++);
	}

	public GenCode() {
		try {
			w = new FileWriter("main.s");
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	void write(String line) {
		try {
			w.write(line);
			w.write("\n");
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	void finish() {
		try {
			w.close();
		} catch (IOException e) {}
	}

	@Override
	public Void visit(IntLiteral il) {
		write("  movq $" + il.n + ", %rax");
		return null;
	}

	@Override
	public Void visit(OpExpr oe) {
		oe.rhs.welcome(this);
		write("  pushq %rax");
		oe.lhs.welcome(this);
		write("  popq %rbx");
		switch (oe.op) {
		case mult: {
			write("  imulq %rbx, %rax");
			break;
		}
		case sub: {
			write("  subq %rbx, %rax");
			break;
		}
		default:
			write("  addq %rbx, %rax");
		}
		return null;
	}

	@Override
	public Void visit(ConDef conDef) {
		write("\n  .globl " + conDef.name + "");
		write("  .type " + conDef.name + ", @function\n");

		write(conDef.name + ":");
		write("  pushq %rbp");
		write("  movq %rsp, %rbp");

		write("  movl	$" + (8 * (1 + conDef.args.size())) + ", %edi");
		write("  call	malloc@PLT");

		write("  movq	$" + conDef.number + ", (%rax)");

    var j = 8;
    for (var i = conDef.args.size() ;i>0;i--) {
			write("  movq	" + (i+1) * 8 + "(%rbp), %rdx");
			write("  movq	%rdx, " + j + "(%rax)");
			j += 8;
		}

		write("  movq %rbp, %rsp");
		write("  popq %rbp");
		write("  ret\n");

		return null;
	}

	@Override
	public Void visit(FunDef funDef) {
		write("\n  .globl " + funDef.name + "");
		write("  .type " + funDef.name + ", @function\n");

		write(funDef.name + ":");

		write("  pushq %rbp");
		write("  movq %rsp, %rbp");

		vars = new HashMap<>();
		var i = funDef.args.size() + 1;
		for (var a : funDef.args)
			vars.put(a.name, (i--) * 8);
		
		i=1;
		for (var a:funDef.localVars)
		  vars.put(a, -((i++)*8));
		write("  subq $"+(8*funDef.localVars.size())+ ", %rsp");
		
		write(".L" + funDef.name + ":");

		funDef.body.welcome(this);
		write("  addq $" + (8 * funDef.localVars.size()) + ", %rsp");

		write("  movq %rbp, %rsp");
		write("  popq %rbp");
		write("  ret\n");

		return null;
	}

	@Override
	public Void visit(Program program) {
		program.cons.forEach(con -> con.welcome(this));
		program.funs.forEach(fun -> fun.welcome(this));
		finish();
		return null;
	}

	@Override
	public Void visit(Variable variable) {
		write("  movq " + vars.get(variable.name) + "(%rbp), %rax");
		return null;
	}

	@Override
	public Void visit(FunCall funCall) {
		funCall.params.forEach((param) -> {
			param.welcome(this);
			write("  pushq %rax");
		});
		if (funCall.isTailRecursive) {
			for (int i = 0; i < funCall.params.size(); i++) {
				write("  popq %rax");
				write("  movq %rax, " + ((2 + i) * 8) + "(%rbp)");
			}
			write("  jmp .L" + funCall.name + "");
		} else {
			write("  call " + funCall.name + "");
			write("  addq $" + (8 * funCall.params.size()) + ", %rsp");
		}
		return null;
	}

	@Override
	public Void visit(Param param) {
		return null;
	}

	@Override
	public Void visit(FieldSelect fieldSelect) {
		fieldSelect.o.welcome(this);
		write("  movq	" + 8 * (1 + fieldSelect.nr) + "(%rax), %rax");
		return null;
	}

	@Override
	public Void visit(IfExpr ifExpr) {
		ifExpr.cond.welcome(this);
		var endLabel = nextLabel();
		var falseLabel = nextLabel();
		write("  cmpq	$0, %rax");
		write("  je	" + falseLabel + "");
		ifExpr.a1.welcome(this);
		write("  jmp	" + endLabel + "");
		write(falseLabel + ":");
		ifExpr.a2.welcome(this);
		write(endLabel + ":");
		return null;
	}
  @Override
  public Void visit(SwitchExpr switchExpr) {
    switchExpr.e.welcome(this);
    write("  movq   (%rax), %rcx");
    var ls = new ArrayList<>();
    switchExpr.cases.forEach(cas->{
      var l = nextLabel();
      write("  cmpq  $"+cas.number+ ", %rcx");
      write("  je  "+l);
      ls.add(l);
    });
    
    var endLabel = nextLabel();
    write("  jmp  "+endLabel);
    
    var it = ls.iterator();
    switchExpr.cases.forEach(cas ->{
      write(it.next()+":");
      cas.welcome(this);
      write("  jmp  "+endLabel);

    });
    write(endLabel+":");
    return null;
  }

  @Override
  public Void visit(Case case1) {
    write("  movq %rax, "+vars.get(case1.var)+"(%rbp)");
    case1.e.welcome(this);
    return null;
  }


}
