package v8.name.panitz.wip;

import java.util.HashSet;
import java.util.List;
import java.util.Set;

public abstract class Tree {
  String type = null;

  abstract <R> R welcome(Visitor<R> v);

  public static class FieldSelect extends Tree {
    Tree o;
    String field;
    int nr = 0;

    public FieldSelect(Tree o, String field) {
      super();
      this.o = o;
      this.field = field;
    }

    @Override
    <R> R welcome(Visitor<R> v) {
      return v.visit(this);
    }
  }

  public static class Param {
    String name;
    String typ;

    public Param(String name, String typ) {
      super();
      this.name = name;
      this.typ = typ;
    }

    String getJBCType() {
      return typ.equals("int") ? "I" : ("L" + typ + ";");
    }

  }

  public static class ConDef extends Tree {
    String name;
    List<Param> args;
    int number;
    static int nr = 0;

    public ConDef(String image, List<Param> args) {
      super();
      this.name = image;
      this.args = args;
      number = nr++;
    }

    @Override
    <R> R welcome(Visitor<R> v) {
      return v.visit(this);
    }

  }

  public static class FunCall extends Tree {
    String name;
    List<Tree> params;
    public boolean isTailRecursive = false;

    public FunCall(String image, List<Tree> params) {
      this.name = image;
      this.params = params;
    }

    @Override
    <R> R welcome(Visitor<R> v) {
      return v.visit(this);
    }

  }

  public static class Variable extends Tree {
    String name;

    public Variable(String name) {
      this.name = name;
    }

    @Override
    <R> R welcome(Visitor<R> v) {
      return v.visit(this);
    }
  }

  public static class Program extends Tree {
    List<ConDef> cons;
    List<FunDef> funs;

    public Program(List<ConDef> cons, List<FunDef> funs) {
      this.cons = cons;
      this.funs = funs;
    }

    @Override
    <R> R welcome(Visitor<R> v) {
      return v.visit(this);
    }
  }

  public static enum Operator {
    mult, add, sub;
  }

  public static class FunDef extends Tree {
    String name;
    List<Param> args;
    Tree body;
    Set<String> localVars;

    public FunDef(String name, List<Param> args, Tree body, String type) {
      super();
      this.name = name;
      this.args = args;
      this.body = body;
      this.type = type;
      this.localVars = new HashSet<>();
    }

    @Override
    <R> R welcome(Visitor<R> v) {
      return v.visit(this);
    }
  }

  public static class IfExpr extends Tree {
    Tree cond;
    Tree a1;
    Tree a2;

    public IfExpr(Tree cond, Tree a1, Tree a2) {
      super();
      this.cond = cond;
      this.a1 = a1;
      this.a2 = a2;
    }

    @Override
    <R> R welcome(Visitor<R> v) {
      return v.visit(this);
    }
  }

  public static class OpExpr extends Tree {
    Tree lhs;
    Operator op;
    Tree rhs;

    public OpExpr(Tree lhs, Operator op, Tree rhs) {
      super();
      this.lhs = lhs;
      this.op = op;
      this.rhs = rhs;
    }

    @Override
    <R> R welcome(Visitor<R> v) {
      return v.visit(this);
    }

  }

  public static class IntLiteral extends Tree {
    int n;

    public IntLiteral(int n) {
      super();
      this.n = n;
    }

    @Override
    <R> R welcome(Visitor<R> v) {
      return v.visit(this);
    }

  }

  public static class SwitchExpr extends Tree {
    Tree e;
    List<Case> cases;
    public SwitchExpr(Tree result, List<Case> result2){
      e=result;
      cases=result2;
    }

    @Override
    <R> R welcome(Visitor<R> v) {
      return v.visit(this);
    }
  }
  public static class Case extends Tree {
    String con;
    String var;
    Tree e;
    public int number;

    public Case(String con, String car, Tree e) {
      super();
      this.con = con;
      this.var = car;
      this.e = e;
    }

    @Override
    <R> R welcome(Visitor<R> v) {
      return v.visit(this);
    }

  }
}
