package v8.name.panitz.wip;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import v8.name.panitz.wip.Tree.Case;
import v8.name.panitz.wip.Tree.ConDef;
import v8.name.panitz.wip.Tree.FieldSelect;
import v8.name.panitz.wip.Tree.FunCall;
import v8.name.panitz.wip.Tree.FunDef;
import v8.name.panitz.wip.Tree.IfExpr;
import v8.name.panitz.wip.Tree.IntLiteral;
import v8.name.panitz.wip.Tree.OpExpr;
import v8.name.panitz.wip.Tree.Param;
import v8.name.panitz.wip.Tree.Program;
import v8.name.panitz.wip.Tree.SwitchExpr;
import v8.name.panitz.wip.Tree.Variable;

public class TypeCheck implements Visitor<String>{
	Map<String, Tree.FunDef> funs;
	Map<String, Tree.ConDef> cons;
  Map<String, String> vars;
  Set<String> localVars;

	@Override
	public String visit(IntLiteral il) {
		il.type="int";
		return "int";
	}

	@Override
	public String visit(OpExpr oe) {
		var lt = oe.lhs.welcome(this);
		var rt = oe.rhs.welcome(this);
		if (!lt.equals("int")) throw new RuntimeException("kein int in Operand. Typ des Operanden: "+lt);
		if (!rt.equals("int")) throw new RuntimeException("kein int in Operand. Typ des Operanden: "+rt);
		oe.type="int";
		return "int";
	}

	@Override
	public String visit(FunDef funDef) {
		vars = new HashMap<>();
		localVars  = new HashSet<>();
		funDef.args.forEach(a -> {
			var t = a.typ;
			vars.put(a.name, t);
			if (!t.equals("Object")&&!t.equals("int")&&cons.get(t)==null) {
				throw new RuntimeException("unknown type: "+t);
			}
		});
		var t = funDef.body.welcome(this);
		if (!t.equals(funDef.type)&&!funDef.type.equals("Object")) {
			throw new RuntimeException("wrong result type for function "+funDef.name+". Declared type: "+funDef.type+" Actual type: "+t);
		}
		funDef.localVars = localVars;
		return t;
	}

	@Override
	public String visit(ConDef conDef) {
		conDef.args.forEach(a -> {
			var t = a.typ;
			if (!t.equals("Object")&&!t.equals("int")&&cons.get(t)==null) {
				throw new RuntimeException("unknown type: "+t);
			}
		});
		conDef.type = conDef.name;
		return conDef.type;
	}

	@Override
	public String visit(Param param) {
		return null;
	}

	@Override
	public String visit(Program program) {
		funs = new HashMap<>();
		program.funs.forEach(fun->funs.put(fun.name, fun));
		cons = new HashMap<>();
		program.cons.forEach(con->cons.put(con.name, con));
		program.cons.forEach(con->con.welcome(this));
		program.funs.forEach(fun->fun.welcome(this));
		
		return "int"; //program.type;
	}

	@Override
	public String visit(Variable variable) {
		var t = vars.get(variable.name);
		if (t==null) {
			throw new RuntimeException("Unkown variable: "+variable.name);
		}
		variable.type = t;
		return t;
	}

	@Override
	public String visit(FunCall funCall) {
		var fun = funs.get(funCall.name);
		if (fun!=null) {
			var args = fun.args;
			checkParameters(funCall, args);
			return funCall.type = fun.type;
		}else {
			var con = cons.get(funCall.name);
			if (con!=null) {
				var args = con.args;
				checkParameters(funCall, args);
			
				return funCall.type = con.type;		
			}
		}
		throw new RuntimeException("unknown function "+funCall.name);
	}

	private void checkParameters(FunCall funCall, List<Param> args) {
		if (args.size()!=funCall.params.size()) {
			throw new RuntimeException("Wrong number of arguments for function "+funCall.name);
		}	
		var as = args.iterator();
		var params = funCall.params.iterator();
		while(as.hasNext()) {
			var arg = as.next();
			var param = params.next();
			var pt = param.welcome(this);
			if (!arg.typ.equals(pt)&&!arg.typ.equals("Object")) {
				throw new RuntimeException("Wrong parameter type for function "+funCall.name);
			}
		}
	}

	@Override
	public String visit(FieldSelect fieldSelect) {
		var ot= fieldSelect.o.welcome(this);
		var con = cons.get(ot);
		if (con==null) throw new RuntimeException("field selection on  non struct type: "+ot);
		
		if (!con.args.stream().anyMatch(arg->arg.name.equals(fieldSelect.field))) {
			throw new RuntimeException("unknown field selection on struct type: "+ot+ " field "+fieldSelect.field);
		}
		var i = 0;
		for (var arg:con.args) {
			if (arg.name.equals(fieldSelect.field)) {
				fieldSelect.nr = i;
				return fieldSelect.type=arg.typ;
			}
			i++;
		}
		return null;
	}

	@Override
	public String visit(IfExpr ifExpr) {
    //TODO: check that both alternatives have same type
		ifExpr.cond.welcome(this);
		ifExpr.a1.welcome(this);
		ifExpr.a2.welcome(this);
		return ifExpr.type=ifExpr.a1.type;
	}
  @Override
  public String visit(SwitchExpr switchExpr) {
    switchExpr.e.welcome(this);
    switchExpr.cases.forEach(cas->switchExpr.type=cas.welcome(this));
    return switchExpr.type;
  }

  @Override
  public String visit(Case case1) {
    if (vars.containsKey(case1.var))
      throw new RuntimeException("variable allready defined: "+case1.var);
    var con = cons.get(case1.con);
    if (con==null)
      throw new RuntimeException("unknown constructor name: "+case1.con);
    localVars.add(case1.var);
    vars.put(case1.var, case1.con);
    case1.number=con.number;
    var result = case1.e.welcome(this);
    vars.remove(case1.var);
    return result;
  }


}
