package v8.name.panitz.wip.antlr;

import java.util.ArrayList;
import java.util.stream.Collectors;

import org.antlr.v4.runtime.misc.Pair;

import v8.name.panitz.wip.antlr.WipBaseListener;
import v8.name.panitz.wip.antlr.WipParser.AddExprContext;
import v8.name.panitz.wip.antlr.WipParser.AddOpContext;
import v8.name.panitz.wip.antlr.WipParser.AtomContext;
import v8.name.panitz.wip.antlr.WipParser.CasesContext;
import v8.name.panitz.wip.antlr.WipParser.ConDefContext;
import v8.name.panitz.wip.antlr.WipParser.ExprContext;
import v8.name.panitz.wip.antlr.WipParser.FunDefContext;
import v8.name.panitz.wip.antlr.WipParser.IfExprContext;
import v8.name.panitz.wip.antlr.WipParser.MultExprContext;
import v8.name.panitz.wip.antlr.WipParser.MultOpContext;
import v8.name.panitz.wip.antlr.WipParser.ParamContext;
import v8.name.panitz.wip.antlr.WipParser.ParamsContext;
import v8.name.panitz.wip.antlr.WipParser.ScaseContext;
import v8.name.panitz.wip.antlr.WipParser.StartSymbolContext;
import v8.name.panitz.wip.antlr.WipParser.SwitchExprContext;
import v8.name.panitz.wip.antlr.WipParser.VarOrFunCallContext;
import v8.name.panitz.wip.Tree.Case;
import v8.name.panitz.wip.Tree.ConDef;
import v8.name.panitz.wip.Tree.FieldSelect;
import v8.name.panitz.wip.Tree.FunCall;
import v8.name.panitz.wip.Tree.FunDef;
import v8.name.panitz.wip.Tree.IfExpr;
import v8.name.panitz.wip.Tree.IntLiteral;
import v8.name.panitz.wip.Tree.OpExpr;
import v8.name.panitz.wip.Tree.Operator;
import v8.name.panitz.wip.Tree.Param;
import v8.name.panitz.wip.Tree.Program;
import v8.name.panitz.wip.Tree.SwitchExpr;
import v8.name.panitz.wip.Tree.Variable;

public class BuildTree extends WipBaseListener{

  @Override
  public void exitStartSymbol(StartSymbolContext ctx) {
    ctx.result = new Program
        (ctx.conDef().stream().map(cc->cc.result).collect(Collectors.toList())
        ,ctx.funDef().stream().map(fc->fc.result).collect(Collectors.toList()));    
  }

  @Override
  public void exitFunDef(FunDefContext ctx) {
    var args = ctx.params().result;
    ctx.result=new FunDef(ctx.Ident(0).getText(), args==null?new ArrayList<>():args, ctx.expr().result, ctx.Ident(1).getText());
  }

  @Override
  public void exitParams(ParamsContext ctx) {
      ctx.result=ctx.param().stream().map(pc->pc.result).collect(Collectors.toList());
  }
  
  @Override
  public void exitParam(ParamContext ctx) {
     ctx.result = new Param(ctx.Ident(0).getText(), ctx.Ident(1).getText());
  }

  @Override
  public void exitExpr(ExprContext ctx) {
    if (null!=ctx.ifExpr()) ctx.result=ctx.ifExpr().result;
    else if (null!=ctx.switchExpr()) ctx.result=ctx.switchExpr().result;
    else if (null!=ctx.addExpr()) ctx.result=ctx.addExpr().result;
  }

  @Override
  public void exitAddExpr(AddExprContext ctx) {
    ctx.result = ctx.multExpr().result;
    for (var p:ctx.addOp()) {
      ctx.result = new OpExpr(ctx.result, p.result.a, p.result.b);
    }
  }
  @Override
  public void exitAddOp(AddOpContext ctx) {
    ctx.result = new Pair<>(ctx.ADD_OP().getText().equals("+")?Operator.add:Operator.sub,ctx.multExpr().result);
  }


  @Override
  public void exitMultExpr(MultExprContext ctx) {
    ctx.result = ctx.atom().result;
    for (var p:ctx.multOp()) {
      ctx.result = new OpExpr(ctx.result, p.result.a, p.result.b);
    }
    
  }

  @Override
  public void exitMultOp(MultOpContext ctx) {
    ctx.result = new Pair<>(Operator.mult,ctx.atom().result);    
  }
  @Override
  public void exitConDef(ConDefContext ctx) {
    var args = ctx.params().result;
    ctx.result = new ConDef(ctx.Ident().getText(), args==null?new ArrayList<>():args);
  }
  @Override
  public void exitAtom(AtomContext ctx) {
    if (ctx.IntegerLiteral()!=null)ctx.result=new IntLiteral(Integer.parseInt(ctx.IntegerLiteral().getText()));
    else if (ctx.expr()!=null) ctx.result = ctx.expr().result; 
    else if (!ctx.varOrFunCall().isEmpty()) ctx.result = ctx.varOrFunCall().result;
  }
  @Override
  public void exitIfExpr(IfExprContext ctx) {
    ctx.result = new IfExpr(ctx.expr(0).result
        , ctx.expr(1).result
        , ctx.expr(2).result);
  }
  
  @Override
  public void exitSwitchExpr(SwitchExprContext ctx) {
    ctx.result=new SwitchExpr(ctx.expr().result,ctx.cases().result);
  }
 
  @Override
  public void exitCases(CasesContext ctx) {
    ctx.result=ctx.scase().stream().map(cc->cc.result).collect(Collectors.toList());
  }
  
  @Override
  public void exitScase(ScaseContext ctx) {
    ctx.result = new Case(ctx.Ident(0).getText(),ctx.Ident(1).getText(),ctx.expr().result);
  }
  @Override
  public void exitVarOrFunCall(VarOrFunCallContext ctx) {
    if (ctx.LPAR()!=null ){
      ctx.result = new FunCall(ctx.Ident(0).getText(), ctx.expr().stream().map(c->c.result).collect(Collectors.toList()) );
    }else {
      ctx.result = new Variable(ctx.Ident(0).getText());
    }
    var it = ctx.Ident().iterator();
    it.next();
    it.forEachRemaining(f->ctx.result = new FieldSelect(ctx.result, f.getText()));     
  }
}
