grammar Wip;
@header {
package v8.name.panitz.wip.antlr;
import static v8.name.panitz.wip.Tree.*;
import v8.name.panitz.wip.Tree;

import java.util.List;
}

startSymbol returns [Tree result]:
   	 (conDef |funDef)*;

funDef returns [FunDef result]:
        'fun' Ident '(' params ')' ':' Ident '=' expr;
    
params returns [List<Param> result]:
	( param (',' param  )* )?;

param returns [Param result]:
    Ident ':' Ident;

conDef returns [ConDef result]:
        'con' Ident '(' params ')';

expr returns [Tree result]:
     ifExpr  
   | switchExpr
   | addExpr  ; 

switchExpr returns [Tree result]:
  'switch' expr cases;
  
cases returns [List<Case> result]:
  ( {$start.getCharPositionInLine()==getCurrentToken().getCharPositionInLine()}?
  	  'case' scase
  )+ ;  

scase returns [Case result]:
  Ident Ident '->' expr ;

ifExpr returns [IfExpr result]:
	'if' expr 'then' expr 'else' expr ;

addExpr returns [Tree result]:
	multExpr (addOp)*;

addOp returns [Pair<Operator,Tree> result]:
	ADD_OP multExpr ;

multExpr returns [Tree result]:
	atom  (multOp)*;

multOp returns [Pair<Operator,Tree> result]:
	'*' atom;	

atom returns [Tree result]:
    IntegerLiteral 
  | LPAR expr ')'
  | varOrFunCall   ;

varOrFunCall returns [Tree result]:
  Ident (LPAR (	expr (',' expr )* )? ')' )?
        ('.' Ident )* ;

ADD_OP : ('+'|'-');

Ident:   Nondigit
        (   Nondigit
        |   Digit
        )*;

IntegerLiteral:    ('0'..'9')+ ;
LPAR: '(';
WS  :   (' ' | '\t' | '\r'| '\n') -> skip;

fragment
Nondigit:   [a-zA-Z_];

fragment
Digit:   [0-9];